from util import *
import anomolies
from operators import *
import pr
import patterns
from specs import SPECIFIERS
from te_gen_util import get_type
import os.path
import config
from domain_model import setup_normalized_transitions
import presentations
import walk
from textentry import group_name_map, do_anomaly
from futil import reads_from, writes_to

class PrintRequest:
    def __repr__ (self):
        return "print request"

class HtmlPR (PrintRequest):
    def __init__(self):
        self.linefeed = '<br>'
        self.start= '<h3>'
        self.end = '</h3>'
        self.focus = '<font color = "104bcl">'
        self.end_focus = '</font>'
        self.intro = '<html> \n'
        self.ending = '</html>'

class TePR (PrintRequest):
    def __init__(self):
        self.linefeed =  ''
        self.start= ''
        self.end = ''
        self.focus = ''
        self.end_focus = ''
        self.intro = ''
        self.ending = ''
        
####################### Record Actions ###########################



def one_new (instance):
    t1 = instance1.get_sc()
    t2 = instances2.get_sc()
    if t1.find('polgen_temp') != -1 or t2.find('polgen_temp') != -1:
        return 1
    else:
        return 0
    
##################### Callbacks #####################################

def adjust_pattern_list_finalCB ():
    """we look at active status of pattern realizations
    and adjust the type contributors
    At this point realizations should have their active property
    set (either by user or by default)"""
    for p in pr.all_pattern_realizations.values():
        patname = the_type(p)[:-1]
        contributor = patname.lower()
        patobj = eval('patterns.' + patname)
        participants = patobj.participants
        #first make sure any type modifications are removed if p is not active
        can_modify_type = patobj.gen_resource_types
        if can_modify_type != 'no':
            for part in participants:
                obj = p.__dict__[part] # the resource object
                if the_type(obj) == 'File' or \
                   (the_type(obj) == 'Process' and can_modify_type == part):
                    if p.active == 0 and contributor in obj.type_contributors:
                        obj.type_contributors.remove(contributor)
                        TY_Replace().go(obj, context2type(obj.original_context))
                    elif p.active == 1 and contributor not in obj.type_contributors:
                        obj.type_contributors.append(contributor)
      
                    
    




def set_integrity_groupCB(process_name, group):
    #print "setting " + process_name + " to group " + group
    resources_dict[process_name].group = group

def set_active_valueCB (item, value):
    v = 1
    if value == 'OFF':
        v = 0 # delete pattern instance
    TY_Adjust_Active().go(pr.pattern_assertions[item], v, "User choice")




def set_file_contextsCB (group, val):
       if val != 'ignore':
        processes = [p for p in resources_dict.values() if the_type(p) == 'Process' and \
                     p.group == group]
        files = []
        if val in ['ALL files read', 'ALL files read or modified']:
            for p in processes:
                #print p
                there = [t.domain for t in p.in_edges  \
                         if the_type(t.domain) == 'File']
                files = files + there
        elif val in ['ALL files modified', 'ALL files read or modified']:
            for p in processes:
                there = [t.range for t in p.out_edges \
                         if the_type(t.range) == 'File']
                files = files + there
        for f in seq2set(files):
            post(f, 'type_contributors', group)
            f.type_contributors.append(group)

def set_scCB (name, value, return_to_original = 0):
    #print "setting " + name + ' to ' + value
    f = resources_dict[name]
    TS.set_sc_for_objCB(f, value)

def set_dir_scCB (name, value):
    obj = resources_dict[name]
    post(obj, 'sc_extent', value)
    obj.sc_extent = value

######################### Set types ###########################################
    


class TypingStrategy:
    def set_sc_for_objCB(self, f, value):
        """called via gui - value determined by user interaction"""
        TY_From_Contributors().go(f, value)
        #TY_Replace().go(f, value)
 





TS = TypingStrategy()





##################### Generators #########################
    



### typing rules
## original context could be set by environment or equal the polgen default
## if the polgen default then it is possible that the type could be changed thru participation in a pattern
##      the new type would be <prgm-name>_<pattern-type>_t
## or thru belonging to a group
##      the new type would be <prgm-name>-<group-name>_t
## environment types are not reset by polgen -
##    that is, we only want to change types for those files that are "owned" by the application
##    if we changed a type from 1_t to 2_t, then every other application that used to communicate
##    with 1_t would now also need to communicate with 2_t - should we allow the policy writer to do this?
### let's say we allow the policy writer to change what he/she wants based on information that polgen provides.
##     two situations
##  1. patterns - writer is notified that if an environmental resource is treated as special there would be a pattern
##  2. groups of read or modified - default is to bring up the original environment set type.  If policy writer chooses
##    to override the default, we pop up a message - " are you sure you want to do this"
##


        
##################### Fc Generation ###############################

def get_dir_extension_to_use (name, extension):
    dir = os.path.dirname(name)
    file_name = name.split('/')[-1].split('.')
    possible_name = ""
    if extension == 'dir/name':
        use = name
    elif len(file_name) > 1:
        possible_name = file_name[0]
        if extension == 'dir/name.*':
            use = dir + '/' + possible_name + '.*'
        else:
            use = dir + extension[3:]
    else: # no . in string
        if extension == 'dir/name.*':
            use = name + '.*'
        else:
            use = dir + extension[3:]
    return use

def do_fc_generation(name_extension_pairs):
    prgm = get_current_program()
    pname = prgm.name
    fileid  = pname + ".fc"
    fileStr = config.result_file(fileid)
    out = open(fileStr, "w")
    fc_strm = open(fileStr, "w")
    print "Writing to " + fileStr
    responses = []
    for item in name_extension_pairs:
        name = item[0]
        obj = resources_dict[name]
        context = obj.get_sc()
        extension = item[1]
        use = get_dir_extension_to_use(name, extension)
        responses.append([use, context])
    responses = seq2set(responses)
    max = 0
    for  r in responses:
        if len(r[0]) > max:
            max = len(r[0])
    for r in responses:
        fname = r[0]
        context = r[1]
        there = len(fname)
        spaces = max - there + 40
        fc_strm.write("%s %*s \n" % (fname, spaces, context))
    fc_strm.close()

################## Te Generation Strategies ########################
    
class GenerationStrategy:

       
    def do_te_generation (self):
        self.open_file()
##        self.do_type_declarations()
##        self.do_allows_and_macros(self.file, 0)
        gen_te_page(self.file, TePR())
        self.open_html_file()
        gen_te_page(self.hfile, HtmlPR())
        self.hfile.close()
        
    def open_file (self):
        prgm = get_current_program()
        pname = prgm.name
        fileid  = pname + ".te"
        fileStr = config.result_file(fileid)
        out = open(fileStr, "w")
        print "Writing to " + fileStr
        self.file = out

    
    def open_html_file (self):
        prgm = get_current_program()
        pname = prgm.name
        fileid  = "te.html"
        fileStr = config.result_file(fileid)
        out = open(fileStr, "w")
        print "Writing to " + fileStr
        self.hfile = out 
        
    def do_type_declarations (self):
        spec_out = self.file
        spec_out.write("##################### Type Declarations #################### \n\n")
        use = []
        found = {}
        for r in resources_dict.values():
            #print r
            orig = r.original_context
            now = r.get_sc()
            if now == None:
                print "security context not set for" + r
            if orig != now:
                if now in found.keys():
                    found[now].append(r)
                else:
                    found[now] = [r]
        for sc in found.keys():
            types = {}
            for support in found[sc]:
                ty = the_type(support).lower()
                if support.special_name != None:
                    ty = support.special_name
                if ty in types.keys():
                    types[ty].append(support)
                else:
                    types[ty] = [support]
            for t in types.keys():
                if t == 'spawn':
                    continue
                elif t in ['file', 'dir']:
                    if sc.find('exec_t') != -1:
                        attribute = 'file_type, exec_type'
                    else:
                        attribute = 'file_type, sysadmfile'
                elif t == 'process':
                    attribute = 'domain'
                else:
                    attribute = t
    
                use.append('type ' + context2type(str(sc)) + ',' \
                                      + attribute + '; \n\n')
        use = seq2set(use)
        for item in use:
            spec_out.write(item)

    
    def close_file(self):
        #print 'closing file'
        self.file.close()
        
    def do_allows_and_macros (self, strm, test_for_coveragep):
        return "done"
    
class NoMacros (GenerationStrategy):
    
    def do_allows_and_macros(self, out, test_for_coveragep = 0): #current with test_for_coverage = 1
        #print normalized_transf.sc_from == 'typegen'
        seen = []
        out.write("\n\n##################### Allow Statements ##################### \n\n\n")
        populate_marks_on_transitions()
        for nt in normalized_trans.values():
            if nt.domain_context.find('polgen_temp') != -1 or nt.range_context.find('polgen_temp') != -1:
                # exists old keys - should not be here
                continue
            #current - only do next if not nt.is_covered()
            #see if there exists a pattern that reports this connection
            if not test_for_coveragep or not nt.is_covered():
                if nt.action != 'spawn':
                    if nt.action not in ['activate', 'exec']:
                        out.write(str(nt) + '\n')
                        out.write('#         because \n')
                        for e in nt.evidence_for:
                            out.write('#            ' + str(e) + '\n')

the_no_macro_strategy = NoMacros()

def populate_marks_on_transitions():
    for realization in [x for x in pr.all_pattern_realizations.values() if x.active == 1]:
        parts = realization.participants
        walk_participants(realization, parts)

def walk_participants (realization, remainder):
    """ remainder is a list of participants """
    if len(remainder) > 1:
        r1 = remainder[0]
        r2 = remainder[1]
        if r2 in trans[r1].keys():
            trans[r1][r2].marked.append(realization)
        elif r1 in trans[r2].keys():
            trans[r2][r1].marked.append(realization)
##        else:
##            print str(realization.pattern) + "(" + str(len(remainder)) + "): " + str(r1) + "--->" + str(r2) + " not in transitions"
        walk_participants(realization, remainder[1:])
#obsolete        
##def gen_pattern_instances2(out, patstr, realization):
##    """ Generate a set of macro statements for the realization"""
##    spec_evidence_dict = {}
##    for individual in walk.instance2individuals(instance):
##            preal = presentations.create_patternR3(patstr, individual)
##            spec = preal.spec()
##            #pattern_assertion_factory(spec,preal, violatorp)
##            if spec in spec_evidence_dict.keys():
##                spec_evidence_dict[spec].append(individual)
##            else:
##                spec_evidence_dict[spec] = [individual]
##    for spec in spec_evidence_dict.keys():
##        out.write(spec)
##        for e in spec_evidence_dict[spec]:
##            out.write('#          because of  ' + str(e[:-1]) + '\n')

def gen_te_page (out, printreq):
    linefeed = printreq.linefeed
    start= printreq.start
    end = printreq.end
    focus = printreq.focus
    end_focus = printreq.end_focus
    intro = printreq.intro
    ending = printreq.ending
    out.write(intro)
    out.write(start + "##################### Type Declarations ####################" \
              + end + " \n\n")
    use = []
    found = {}
    for r in resources_dict.values():
        orig = r.original_context
        now = r.get_sc()
        if now == None:
            print "security context not set for" + r
        if orig != now and now.find(PRGM.name) != -1:
            if now in found.keys():
                found[now].append(r)
            else:
                found[now] = [r]
    for sc in found.keys():
        types = {}
        for support in found[sc]:
            ty = the_type(support).lower()
            if support.special_name != None:
                ty = support.special_name
            if ty in types.keys():
                types[ty].append(support)
            else:
                types[ty] = [support]
        for t in types.keys():
            if t == 'spawn':
                continue
            elif t in ['file', 'dir']:
                if sc.find('exec_t') != -1:
                    attribute = 'file_type, exec_type'
                else:
                    attribute = 'file_type, sysadmfile'
            elif t == 'process':
                attribute = 'domain'
            else:
                attribute = t

            use.append('type ' + context2type(str(sc)) + ',' \
                                  + attribute + '; \n\n')
    use = seq2set(use)
    for item in use:
        out.write(linefeed + focus + item + end_focus + ' \n')
    out.write(start + "##################### Allow Statements ####################" + end + "\n\n")
    for nt in normalized_trans.values():
        if nt.domain_context.find('polgen_temp') != -1 \
           or nt.range_context.find('polgen_temp') != -1:
                # exists old keys - should not be here
                continue
        #see if there exists a pattern that reports this connection
        if not nt.is_covered():
            if nt.action not in ['spawn','activate', 'exec']:
                out.write(focus + linefeed + str(nt) + end_focus + ' \n')
                out.write(linefeed + '#--------------because \n')
                for e in nt.evidence_for:
                    out.write(linefeed + '#-----------------' + str(e.name) + '\n')
    out.write(start + "##################### Pattern Instances ####################" + end + "\n\n")
    instances = [pa for pa in pr.pattern_assertions_final.values() if pa.accepted == 1]
    for patstr in patterns.PATTERN_TYPES:
        current = [pa for pa in instances if pa.pattern.name == patstr]
        if len(current) != 0:
            for v in current:
                out.write(linefeed + focus + pr.all_pattern_realizations[v.evidence[0]].spec() \
                           + end_focus + '\n')
                #strm.write(v.print_form)
                for e in v.evidence: #set of realizations:
                        out.write(linefeed + '#--------------------' + \
                                  str([x.name for x in pr.all_pattern_realizations[e].constellation[:-1]]) + '\n')
                out.write(' \n')
    out.write(ending)

class WithMacros (GenerationStrategy):
##    def do_allows_and_macros(self, strm, test_for_coveragep):
##        the_no_macro_strategy.do_allows_and_macros(strm, 1)
##        strm.write("##################### Pattern Instances #################### \n\n")
##        for patstr in patterns.PATTERN_TYPES:
##            instances = []
##            recogs = [recog for recog in SPECIFIERS if recog.pattern.name == patstr]
##            for recog in recogs:
##                if len(PRGM.pattern_instances[recog.__name__]) != 0:
##                    #strm.write('#        ' + recog.pattern.name + ' instances observed by ' + recog.__name__ + '\n')
##                    instances = instances + PRGM.pattern_instances[recog.__name__]
##            for v in instances:
##                gen_pattern_instances2(strm, patstr, v)
##                strm.write(' \n')
    def do_allows_and_macros(self, strm, test_for_coveragep): #this is current stategy
        the_no_macro_strategy.do_allows_and_macros(strm, 1)
        strm.write("##################### Pattern Instances #################### \n\n")
        instances = [pa for pa in pr.pattern_assertions_final.values() if pa.accepted == 1]
##        print "****"
####        print len(pr.pattern_assertions)
##        print instances[0].pattern
        for patstr in patterns.PATTERN_TYPES:
            #print patstr
            current = [pa for pa in instances if pa.pattern.name == patstr]
            #print len(current)
            if len(current) != 0:
                for v in current:
                    strm.write(pr.all_pattern_realizations[v.evidence[0]].spec() + '\n')
                    #strm.write(v.print_form)
                    for e in v.evidence: #set of realizations:
                        strm.write('#          because of  ' + \
                                   str(pr.all_pattern_realizations[e].constellation[:-1]) + '\n')
                    strm.write(' \n')



class WithMacrosForSurvivorsOnly (GenerationStrategy):
    def do_allows_and_macros(self):
        return "done"

gen_strategy = WithMacros()

def clean_up ():
    for r in resources_dict.values():
        new_sc = temp_type_replacement(r)
        if new_sc!= r.get_sc():
            TY_Replace().go(r, context2type(new_sc))
        domain = PRGM.unconfined_domain
        if domain != 'unconfined_t':
            #need to change all unconfined_t
            pos = context2type(new_sc).find('unconfined_t')
            if pos != -1:
                TY_Replace().go(r, domain, 'setting unconfined_domain')
            
    setup_normalized_transitions(1) # re-do with new types

def extending_super_capabilities (process):
    """ written for cgi scripts may need to extend"""
    toreturn = 0
    cgis = [preal for preal in pr.all_pattern_realizations.values() \
            if preal.pattern.name == 'CGI']
    for cgi in cgis:
        if process == cgi.script:
            super = cgi.server
            super_from = seq2set([x.security_context for x in reads_from(super)])
            super_to = seq2set([x.security_context for x in writes_to(super)])
            p_from = seq2set([x.security_context for x in reads_from(process)])
            p_to = seq2set([x.security_context for x in writes_to(process)])
            from_diff = setdiff(p_from, super_from)
            to_diff = setdiff(p_to, super_to)
            if len(from_diff) > 0 or len(to_diff) > 0:
                 toreturn = 1
            break
    return toreturn
        
def do_policy_generationCB():
    for p in [x for x in resources_dict.values() if the_type(x) == 'Process']:
            if p.type_contributors != []:
                ty = final_stype(p, "all")
                if extending_super_capabilities(p):
                    print "Changing type of " + p.name
                    u_and_r = user_and_role(p.security_context)
                    p.security_context = u_and_r + ':' + ty
    items = [f for f in resources_dict.values() if the_type(f) == 'File']
##    for f in resources_dict.values():
##        if f.get_sc() == None:
##            set_sc_final(f)
    data = [[f.name, f.sc_extent] for f in items if \
            f.original_context != f.get_sc() \
            or f.sc_from == 'typegen']
            #f.original_context != f.get_sc()]

    problems = anomolies.multiple_transitions()
    if len(problems) > 0:
        do_anomaly()
        print "Repairing multiple transition anomalies"
        for k in problems.keys():
            TY_Repair_Multiple_Trans().go(problems[k])
    clean_up()
    redo_pattern_assertions()
    do_fc_generation(data) 
    gen_strategy.do_te_generation()

def redo_pattern_assertions():
    for preal in pr.all_pattern_realizations.values():
        #print preal
        pr.pattern_assertion_final_factory(preal)

#dead code   
def set_sc_final(f):
    #only called when no_gui
    """default strategy when user has not intervened"""
    if not PRGM.unlimited_typing or \
       (not f.sc_from == 'typegen' and not PRGM.outsiders_mutable):
        f.set_sc(f.original_context, 'high')
    else: # use contributors
        u_and_r = user_and_role(f.original_context)
        there = f.get_sc()
        fst = final_stype(f, "all")
        fsc = u_and_r + fst
        if there != fsc:
            #print "---- " + there
            TY_Replace().go(f, fst)
            #print "***" + f.get_sc()


